#!/bin/bash
# init script for frontier-squid, runs as unprivileged user

. /etc/init.d/functions

RETVAL=0

CONF_DIR=/etc/squid
SBIN_DIR=/usr/sbin
AWSTATS_SCRIPT=/etc/awstats/run_awstats.sh
SQUID_COMPRESS_LOGS=${SQUID_COMPRESS_LOGS:-true}
DEFAULT_MAX_ACCESS_LOG=1G
if [ "$SQUID_COMPRESS_LOGS" = true ]; then
    DEFAULT_MAX_ACCESS_LOG=5G
fi
#size at which rotateiflarge will rotate access.log
#can be in bytes, or if ending 'M' in megabytes, or if ending 'G' in gigabytes
#use LARGE_ACCESS_LOG if set for backward compatibility
SQUID_MAX_ACCESS_LOG=${SQUID_MAX_ACCESS_LOG:-${LARGE_ACCESS_LOG:-$DEFAULT_MAX_ACCESS_LOG}}
case $SQUID_MAX_ACCESS_LOG in
    *M) let SQUID_MAX_ACCESS_LOG="${SQUID_MAX_ACCESS_LOG%M}*1000000";;
    *G) let SQUID_MAX_ACCESS_LOG="${SQUID_MAX_ACCESS_LOG%G}*1000000000";;
esac
#Set to false to not clean ufs cache on start.  Note however that it will
#  cause swap.state files to grow without bounds if doing log compression.
SQUID_CLEAN_CACHE_ON_START=${SQUID_CLEAN_CACHE_ON_START:-true}

if [ -z "$SQUID_NUM_SERVICES" ]; then
 SQUID_NUM_SERVICES=1
fi
if [ "$SQUID_NUM_SERVICES" -gt 1 ]; then
 SQUIDLABEL="$SQUID_NUM_SERVICES Frontier Squids"
 PLURAL="s"
else
 SQUIDLABEL="Frontier Squid"
 PLURAL=""
fi

CONF_FILE=$CONF_DIR/squid.conf

if [ `id -u` = 0 ]; then
  echo "ERROR: `basename $0` not allowed to be run as root" >&2
  exit 1
fi

if [ ! -d "$CONF_DIR" ] || [ ! -w "$CONF_DIR" ]; then
 echo "ERROR: $CONF_DIR does not exist or is not writable" >&2
 exit 1
fi

if ${SQUID_CUSTOMIZE:-true}; then
 # see if squid.conf needs to be regenerated
 if [ $CONF_FILE.frontierdefault -nt $CONF_FILE ] ||
	 [ $CONF_DIR/customize.sh -nt $CONF_FILE ] ||
	 [ $0 -nt $CONF_FILE ]; then
  if [ -f "$CONF_FILE" ] && [ -n "`find "$CONF_FILE" -perm -0200 2>/dev/null`" ]; then
   echo "ERROR: $CONF_FILE is writable and was about to be overwritten" >&2
   echo "  This may indicate that squid was upgraded from an older version" >&2
   echo "  that was not based on customize.sh." >&2
   echo "  Remove or rename it, or chmod it to be non-writable." >&2
   echo "  To not use customize.sh, see frontier-squid documentation on SQUID_CUSTOMIZE." >&2
   exit 1
  fi
  echo "Generating $CONF_FILE"
  rm -f $CONF_FILE.new
  if $CONF_DIR/customize.sh <$CONF_FILE.frontierdefault >$CONF_FILE.new; then
   rm -f $CONF_FILE.old
   if [ -f $CONF_FILE ]; then
    echo "Saving old configuration as $CONF_FILE.old"
    mv $CONF_FILE $CONF_FILE.old
   fi
   mv $CONF_FILE.new $CONF_FILE
   chmod 444 $CONF_FILE
  else
   exit $?
  fi
 fi
fi

# assume there are no spaces in these file paths
CACHE_DIR=`awk '$1 == "cache_dir" {x=$3} END{print x}' $CONF_FILE`
CACHE_TYPE=`awk '$1 == "cache_dir" {x=$2} END{print x}' $CONF_FILE`
CACHE_LOG=`awk '$1 == "cache_log" {x=$2} END{print x}' $CONF_FILE`
WORKERS=`awk '$1 == "workers" {x=$2} END{print x}' $CONF_FILE`
ACCESS_LOG=`awk '$1 == "access_log" {x=$2} END{print x}' $CONF_FILE`
PID_FILENAME=`awk '$1 == "pid_filename" {x=$2} END{print x}' $CONF_FILE`

if [ -z "$WORKERS" ]; then
  WORKERS=1
fi

expandsmpmacros()
{
 typeset VAR=
 typeset SVAR=
 typeset SERVICE_NAME=squid
 typeset S=0
 typeset VARS=""
 typeset N
 while [ $S -lt $SQUID_NUM_SERVICES ]; do
  if [ "$SQUID_NUM_SERVICES" -gt 1 ]; then
   SERVICE_NAME=$S
  fi
  let S+=1
  SVAR="${1//\$\{service_name\}/$SERVICE_NAME}"
  N=0
  while [ $N -lt $WORKERS ]; do
   let N+=1
   VAR="${SVAR//\$\{process_name\}/${SERVICE_NAME}$N}"
   VAR="${VAR//\$\{process_number\}/$N}"
   if [[ " $VARS " != *" $VAR "* ]]; then
    # not yet in the list, add it
    if [ -n "$VARS" ]; then
     VARS="$VARS "
    fi
    VARS="$VARS$VAR"
   fi
  done
 done
 echo "$VARS"
}

if [ -z "$CACHE_DIR" ]; then
 echo "ERROR: cache_dir not found in $CONF_FILE" >&2
 exit 1
fi
if [ "$CACHE_DIR" == "/" ]; then
 echo "ERROR: cache_dir not allowed to be '/' in $CONF_FILE" >&2
 exit 1
fi
for D in `expandsmpmacros $CACHE_DIR`; do
 if [ ! -d "$D" ]; then
  if [ ! -w "`dirname $D`" ]; then
   echo "ERROR: cache_dir $D does not exist and its parent is not writable" >&2
   exit 1
  fi
  mkdir $D
 elif [ ! -w "$D" ]; then
  echo "ERROR: cache_dir $D is not writable" >&2
  exit 1
 fi
done
if [ $WORKERS -gt 1 ] && [ "$CACHE_TYPE" != rock ]; then
 case "$CACHE_DIR" in
  *\${process_number}*|*\${process_name}*) ;;
  *) echo 'ERROR: when workers > 1, cache_dir must either be rock type or contain ${process_number} or ${process_name}' >&2
     exit 1
     ;;
 esac
 COLLAPSED_FORWARDING=`awk '$1 == "collapsed_forwarding" {x=$2} END{print x}' $CONF_FILE`
 if [ "$COLLAPSED_FORWARDING" = on ]; then
  MEMORY_CACHE_SHARED=`awk '$1 == "memory_cache_shared" {x=$2} END{print x}' $CONF_FILE`
  if [ "$MEMORY_CACHE_SHARED" != "off" ]; then
   echo 'ERROR: when workers > 1 and collapsed_forwarding is on, cache_dir must either be rock type or memory_cache_shared must be off - see http://bugs.squid-cache.org/show_bug.cgi?id=4311' >&2
   exit 1
  fi
 fi
fi
if [ -z "$CACHE_LOG" ]; then
 echo "ERROR: cache_log not found in $CONF_FILE" >&2
 exit 1
fi
CACHE_LOG_DIR=`dirname "$CACHE_LOG"|sed 's/daemon://'`
for D in `expandsmpmacros $CACHE_LOG_DIR`; do
 if [ ! -d "$D" ]; then
  if [ ! -w "`dirname $D`" ]; then
   echo "ERROR: cache_log directory $D does not exist and its parent is not writable" >&2
   exit 1
  fi
  mkdir $D
 elif [ ! -w "$D" ]; then
  echo "ERROR: cache_log directory $D is not writable" >&2
  exit 1
 fi
done
if [ -z "$ACCESS_LOG" ]; then
 echo "ERROR: access_log not found in $CONF_FILE" >&2
 exit 1
fi
if [ "$ACCESS_LOG" != "none" ]; then
 ACCESS_LOG_DIR=`dirname "$ACCESS_LOG"|sed 's/daemon://'`
 if [ "$ACCESS_LOG_DIR" != "$CACHE_LOG_DIR" ]; then
  echo "ERROR: access_log directory $ACCESS_LOG_DIR is not the same as the cache_log directory $CACHE_LOG_DIR" >&2
  exit 1
 fi
else
 ACCESS_LOG_DIR="$CACHE_LOG_DIR"
fi
if [ -z "$PID_FILENAME" ]; then
 echo "ERROR: pid_filename not found in $CONF_FILE" >&2
 exit 1
fi
PID_FILENAME_DIR=`dirname "$PID_FILENAME"`
for D in `expandsmpmacros $PID_FILENAME_DIR`; do
 if [ ! -d "$D" ] || [ ! -w "$D" ]; then
  echo "ERROR: pid_filename directory $D does not exist or is not writable" >&2
  exit 1
 fi
done
if [ $SQUID_NUM_SERVICES -gt 1 ]; then
 for V in $CACHE_DIR $CACHE_LOG_DIR $PID_FILENAME; do
  case "$V" in
    *\${service_name}*) ;;
    *) echo 'ERROR: when $SQUID_NUM_SERVICES > 1 then cache_dir, cache_log, access_log, and pid_filename must all contain ${service_name}' >&2
       exit 1;;
  esac
 done
fi

LOGFILE_ROTATE=`awk '$1 == "logfile_rotate" {x=$2} END{print x}' $CONF_FILE`
if [ -z "$LOGFILE_ROTATE" ]; then
 LOGFILE_ROTATE=10
fi

SQUIDEXE=$SBIN_DIR/squid
if [ $SQUID_NUM_SERVICES -eq 1 ]; then
 SQUID=$SQUIDEXE
else
 SQUID=runsquids
fi

runsquids()
{
 let N=0
 while [ $N -lt $SQUID_NUM_SERVICES ]; do
  $SQUIDEXE -n $N "$@" &
  sleep 2
  let N+=1
 done
 wait
}

isrunning()
{
 # return 0 if all running, else 1
 for F in `expandsmpmacros $PID_FILENAME`; do
  if [ ! -f "$F" ]; then
   return 1
  fi
  typeset PID="`cat <$F 2>/dev/null`"
  if [ ! -n "$PID" ]; then
   rm -f "$F"
   return 1
  fi
  # Need to make sure that $PID actually refers to squid and not some
  #   other program after a system failure & restart.
  # Can't use pidof, because nightly prelink confuses it when $SBIN_DIR
  #   is a system directory.
  # Can't use ls -iL on /proc/$PID/exe, because that shows the inode
  #   of the executable before prelink messed with it, so instead extract
  #   the name of the symlink target and do -iL on that.
  typeset PIDEXE="`ls -l /proc/$PID/exe 2>/dev/null|sed 's,.*-> ,,'`"
  if [ -n "$PIDEXE" ]; then
   # Compare the inodes of $PID's executable and $SQUID to see if they're
   #  the same file.
   # Note that the syntax ${VAR%% *} removes everything in $VAR after a blank.
   # RHEL6-based systems can have a " (deleted)" as part of $PIDEXE.
   typeset PIDINODE="`ls -iL ${PIDEXE%% *} 2>/dev/null`"
   typeset SQUIDINODE="`ls -iL ${SQUIDEXE%% *}`"
   if [ "${PIDINODE%% *}" = "${SQUIDINODE%% *}" ]; then
    continue
   fi
  fi
  return 1
 done
 return 0
}

cleancache()
{
 for D in `expandsmpmacros $CACHE_DIR`; do
  if [ -f $D/swap.state ] || [ -f $D/rock ]; then
   # a cache exists
   echo "Checking if $SQUIDLABEL running..."
   # ignore stderr because squid-3 insists on enabling debugging with -k parse
   if $SQUID -k parse 2>/dev/null; then
    if isrunning; then
     echo "Squid$PLURAL running, doing nothing"
     exit 1
    fi
   else
    RETVAL=$?
    # now run check just to get the error message out; it doesn't enable debug
    $SQUID -k check
    echo "ERROR parsing $CONF_FILE" >&2
    exit $RETVAL
   fi
   echo "Start clearing Cache$PLURARL of $SQUIDLABEL in the background... "
   $SQUID -k shutdown 2>/dev/null||true  # in case all are not shut down
   for D in `expandsmpmacros $CACHE_DIR`; do
    DELD=$D/.deleting/$$
    rm -rf $DELD
    mkdir -p $DELD
    mv $D/* $DELD 2>/dev/null
   done
   nohup $SHELL -c "for D in `expandsmpmacros $CACHE_DIR`; do
    rm -rf \$D/.deleting
   done" </dev/null >/dev/null 2>&1 &
   echo "started"
   break
  fi
 done
 echo "Initializing Cache${PLURAL}..."
 $SQUID -z -d 0
}

removecache()
{
 if isrunning; then
  echo "Squid$PLURAL running, doing nothing"
  exit 1
 fi
 for D in `expandsmpmacros $CACHE_DIR`; do
  echo "Removing cache in $D..."
  rm -rf $D/* $D/.??*
 done
}

start_squid()
{
 echo "Starting $SQUIDLABEL... "
 $SQUID
 # Allow some time for squid initialization to fail.  Also, background
 #  cache cleaning can cause the start to be quite slow and it might take
 #  extra time for the executables to start when using multiple squids.
 sleep 10
 isrunning
 RETVAL=$?
 if [ $RETVAL != 0 ] ; then
  echo "Frontier Squid start failed!!!" >&2
  $SQUID -k shutdown 2>/dev/null || true  # in case all are not shut down
 else
  sleep 2
  echo "done"
 fi

 [ $RETVAL == 0 ]
}

stop_squid()
{
 echo -n "Stopping $SQUIDLABEL... "
 $SQUID -k shutdown
 RETVAL=$?
 if [ $RETVAL != 0 ] ; then
  echo "Frontier Squid stop failed!!!" >&2
 else
  sleep 35 # wait for the default shutdown_lifetime and a little more
  echo "done"
 fi
 [ $RETVAL == 0 ]
}

hasawstats()
{
 [ -n "$AWSTATS_SCRIPT" ] && [ -x "$AWSTATS_SCRIPT" ] && [ "$ACCESS_LOG" != none ] && [ -z "" ]
}

lockrotate()
{
 typeset TMPLOCK
 ROTATELOCKFILE=`expandsmpmacros $ACCESS_LOG_DIR|awk '{print $1}'`/.rotatelock
 TMPLOCK="$ROTATELOCKFILE.$$"
 trap "rm -f $TMPLOCK" 0
 echo "$$" >$TMPLOCK
 if ! ln $TMPLOCK $ROTATELOCKFILE 2>/dev/null; then
  typeset OTHERPID
  OTHERPID="`cat $ROTATELOCKFILE`"
  if [ -z "$OTHERPID" ] || ! kill -0 "$OTHERPID" 2>/dev/null; then
   rm -f $ROTATELOCKFILE
   if ! ln $TMPLOCK $ROTATELOCKFILE 2>/dev/null; then
    echo "Attempted to break lock $ROTATELOCKFILE from $OTHERPID but failed, giving up"
    exit 1
   fi
   echo "Broke lock $ROTATELOCKFILE from $OTHERPID because process was not running"
  else
   echo "Cannot create $ROTATELOCKFILE, already created by running process $OTHERPID"
   exit 1
  fi
 fi
 rm -f $TMPLOCK
 trap 0
}

unlockrotate()
{
 rm -f $ROTATELOCKFILE
}

callawstats()
{
 for LDIR in `expandsmpmacros $ACCESS_LOG_DIR`; do
  typeset LFILE=$LDIR/`basename $ACCESS_LOG`
  typeset AWSTATS_CMD="${AWSTATS_SCRIPT} ${LFILE}$1"
  echo "Calling ${AWSTATS_CMD}"
  ${AWSTATS_CMD}
 done
}

rotate()
{
 # need to first make sure no other rotate is running
 lockrotate
 echo "Rotating $SQUIDLABEL Logs... "
 if $SQUID_COMPRESS_LOGS; then
  typeset DELAYCOMPRESS=""
  typeset STARTRM="0"
  typeset STARTN="0"
  typeset ROTATEN="$LOGFILE_ROTATE"
  if hasawstats; then
   # when awstats script exists, don't compress the first file
   #  because it needs to parse that file
   DELAYCOMPRESS="delaycompress"
   STARTRM="1"
  else
   # in order to look like standard logrotate, start suffix numbers from 1
   #  and save one less file
   STARTN="1"
   let ROTATEN-=1
  fi
  typeset TMPF="`mktemp`"
  trap "rm -f $TMPF*" 0
  for D in `expandsmpmacros $ACCESS_LOG_DIR`; do
   cat >$TMPF <<!EOF!
    $D/access.log $D/cache.log $D/store.log {
     start $STARTN
     compress
     copytruncate
     daily
     $DELAYCOMPRESS
     missingok
     rotate $ROTATEN
    }
!EOF!
   ionice -n7 logrotate -f -s $TMPF.s $TMPF
   rm -f $TMPF*
   # remove non-compressed versions in case switched over
   rm -f $D/*.log.[$STARTRM-9] $D/*.log.[1-9][0-9] $D/*.log.[1-9][0-9][0-9]
  done
  if $SQUID_CLEAN_CACHE_ON_START; then
   # truncate ufs swap.state files because without doing squid -k rotate,
   #  swap.state grows without bounds.  It is only used for restarting with
   #  a warm cache, and we normally delete the cache at startup anyway.
   for D in `expandsmpmacros $CACHE_DIR`; do
    if [ -s $D/swap.state ]; then
     > $D/swap.state
    fi
   done
  fi
  trap 0
 else
  # don't compress
  typeset ROTATESUFFIX
  let ROTATESUFFIX="$LOGFILE_ROTATE - 1"
  for D in `expandsmpmacros $ACCESS_LOG_DIR`; do
   # delete the log files that will be kicked out first, because
   #  otherwise squid stops serving traffic while it removes them
   rm -f $D/access.log.$ROTATESUFFIX $D/cache.log.$ROTATESUFFIX $D/store.log.$ROTATESUFFIX
   # remove compressed versions in case switched over
   rm -f $D/*.log.[1-9]*.gz
  done
  $SQUID -k rotate
  RETVAL=$?
 fi
 if [ $RETVAL != 0 ]; then
  echo "failed"
 else
  echo "done"
 fi
 if hasawstats; then
  callawstats .0
 fi
 unlockrotate
 [ $RETVAL == 0 ]
}

rotateiflarge()
{
 if [ "$ACCESS_LOG" = none ]; then
  echo "No access log to rotate"
  return
 fi
 # if any access.log is too large, rotate them all
 for LDIR in `expandsmpmacros $ACCESS_LOG_DIR`; do
  LFILE=$LDIR/`basename $ACCESS_LOG`
  if [ -f "$LFILE" ] && [ "$(stat -c %s $LFILE)" -gt "$SQUID_MAX_ACCESS_LOG" ]; then
    rotate
    return
  fi
 done
 echo "It's not necessary to rotate the Frontier Squid Logs"
 if hasawstats; then
  # first make sure no other rotate or awstats is running
  lockrotate
  callawstats
  unlockrotate
 fi
}

reload()
{
 echo "Reloading Frontier Squid configuration"
 $SQUID -k reconfigure
}

start()
{
 if ! $SQUID_CLEAN_CACHE_ON_START || cleancache; then
  start_squid
 fi
 [ $RETVAL == 0 ]
}

stop()
{
 stop_squid
 [ $RETVAL == 0 ]
}

restart()
{
 if stop; then
  sleep 3
  start
  return $?
 else
  return $?
 fi
}

case "$1" in
  start)
        start
	exit $?
        ;;
  stop)
        stop
	exit $?
        ;;
  status)
        if isrunning; then
	    echo "$SQUIDLABEL running"
	    exit 0
	fi 
	echo "$SQUIDLABEL not running"
        exit 1
        ;;
  restart)
	restart
	exit $?
        ;;
  reload)
	reload
        exit $?
	;;
  condrestart)
	if isrunning; then
	    restart
	    exit $?
	else
	    echo "$SQUIDLABEL not running"
  	fi
        ;;
  cleancache)
	cleancache
	exit $?
	;;
  removecache)
	removecache
	exit $?
	;;
  rotate)
	rotate
	exit $?
	;;
  rotateiflarge)
	rotateiflarge
	exit $?
	;;
  rebuildconf)
	# already done above
        ;;
  *)
        echo $"Usage: $0 {start|stop|restart|condrestart|reload|status|cleancache|removecache|rotate|rotateiflarge}"
        exit 1
esac

